from flask import Flask, jsonify, render_template, request, make_response
from flask_caching import Cache

import polars as pl
import numpy as np

from hkjc.live import live_odds, _fetch_live_races
from hkjc.harville_model import fit_harville_to_odds
from hkjc.historical import get_horse_data
from hkjc.speedpro import speedmap, speedpro_energy
from hkjc.strategy import qpbanker, place_only
from hkjc import generate_all_qp_trades, generate_all_pla_trades, pareto_filter


def arr_to_dict(arr: np.ndarray, dtype=float):
    """Convert 0-indexed numpy array into 1-indexed nested dictionary

    Args:
        arr (np.ndarray): 0-indexed numpy array
        dtype (type, optional): data type. Defaults to float.

    Returns:
        dict: 1-indexed nested dictionary
    """
    if arr.ndim == 1:
        return {i+1: dtype(np.round(v, 1)) for i, v in enumerate(arr) if not (np.isnan(v) or np.isinf(v))}

    return {i+1: arr_to_dict(v) for i, v in enumerate(arr)}


app = Flask(__name__)
config = {
    "CACHE_TYPE": "RedisCache",
    "CACHE_REDIS_HOST": "localhost",
    "CACHE_REDIS_PORT": "6379"
}
app.config.from_mapping(config)
cache = Cache(app)


@app.route('/')
def disp_race_info():
    race_info = _fetch_live_races('', '')

    try:
        df_speedpro = speedpro_energy(race_info['Date'])
        for race_num, race in race_info['Races'].items():
            for i, runner in enumerate(race['Runners']):
                df = (df_speedpro
                      .filter(pl.col('RaceNo') == race_num)
                      .filter(pl.col('RunnerNumber') == int(runner['No']))
                      )
                race_info['Races'][race_num]['Runners'][i]['SPEnergy'] = df['SpeedPRO_Energy_Difference'].item(
                    0)
                race_info['Races'][race_num]['Runners'][i]['Fitness'] = df['FitnessRatings'].item(
                    0)
    except:  # fill with dummy value if SpeedPro not available
        for race_num, race in race_info['Races'].items():
            for i, runner in enumerate(race['Runners']):
                race_info['Races'][race_num]['Runners'][i]['SPEnergy'] = 0
                race_info['Races'][race_num]['Runners'][i]['Fitness'] = 0

    return render_template('index.html',
                           race_info=race_info)


turf_going_dict = {'FIRM': 'F',
                   'GOOD TO FIRM': 'GF',
                   'GOOD': 'G',
                   'GOOD TO YIELDING': 'GY',
                   'YIELDING': 'Y',
                   'YIELDING TO SOFT': 'YS',
                   'SOFT': 'S',
                   'HEAVY': 'H'}
aw_going_dict = {'WET FAST': 'WF',
                 'FAST': 'FT',
                 'GOOD': 'GD',
                 'SLOW': 'SL',
                 'WET SLOW': 'WS',
                 'RAIN AFFECTED': 'RA',
                 'NORMAL WATERING': 'NW'}
going_dict = {'TURF': turf_going_dict, 'ALL WEATHER TRACK': aw_going_dict}


@app.route('/horse_info/<horse_no>', methods=['GET'])
# cache horse history for 1 day
@cache.cached(timeout=86400, query_string=True)
def disp_horse_info(horse_no):
    # read optional filters
    dist = request.args.get('dist', type=int)
    track = request.args.get('track')
    going = request.args.get('going')
    venue = request.args.get('venue')

    if track not in going_dict.keys():
        track = None
    if venue not in ['HV', 'ST']:
        venue = None
    if (going is not None) and (track is not None) and (going in going_dict[track].keys()):
        going = going_dict[track][going]  # translate going to code
    else:
        going = None

    df = get_horse_data(horse_no)

    if df.height > 0:
        if dist is not None:
            df = df.filter(pl.col('Dist') == dist)
        if track and track.upper() == 'TURF':
            df = df.filter(pl.col('Track') == 'Turf')
        elif track and track.upper() == 'ALL WEATHER TRACK':
            df = df.filter(pl.col('Track') == 'AWT')
        if going is not None:
            df = df.filter(pl.col('G').str.starts_with(going[0]))
        if venue is not None:
            df = df.filter(pl.col('Venue') == venue)

    return render_template('horse-info.html', df=df.head(5))


@app.route('/live_odds/<int:race_no>')
def disp_live_odds(race_no=1):
    odds_dict = live_odds('', '', race_no)
    fitted_odds = fit_harville_to_odds(odds_dict)['odds_fit']
    odds_json = {'Raw': {k: arr_to_dict(v) for k, v in odds_dict.items()},
                 'Fit': {k: arr_to_dict(v) for k, v in fitted_odds.items()}
                 }

    return jsonify(odds_json)


@app.route('/speedmap/<int:race_no>')
def disp_speedmap(race_no=1):
    return speedmap(race_no)


@app.route('/qp/<int:race_no>/<int:banker>/<cover>', methods=['GET'])
def disp_qp_metrics(race_no=1, banker=1, cover='2'):
    use_filter = request.args.get('filter')
    use_filter = use_filter and (use_filter.lower() == 'true')

    odds = live_odds('', '', race_number=race_no)

    if use_filter:
        res = fit_harville_to_odds(odds)
        if res['success']:
            odds['PLA'] = res['odds_fit']['PLA']
    covered = [int(v) for v in cover.split(',') if int(v) != banker]
    ev = qpbanker.expected_value(odds['PLA'], odds['QPL'], banker, covered)
    win = qpbanker.win_probability(odds['PLA'], banker, covered)
    avg_odds = qpbanker.average_odds(odds['QPL'], banker, covered)

    return {'Banker': banker, 'Covered': covered, 'ExpValue': round(ev, 2), 'WinProb': round(win, 2), 'AvgOdds': round(avg_odds, 2)}

@app.route('/pla/<int:race_no>/<cover>')
def disp_pla_metrics(race_no=1, cover=[]):
    odds = live_odds('', '', race_number=race_no)

    res = fit_harville_to_odds(odds)
    if not res['success']:
        raise RuntimeError(
            f"[ERROR] Harville model fitting failed: {res.get('message','')}")
    
    covered = [int(v) for v in cover.split(',')]
    ev = place_only.expected_value(odds['PLA'], res['prob_fit']['P'], covered)
    win = place_only.win_probability(res['prob_fit']['P'], covered)
    avg_odds = place_only.average_odds(odds['PLA'], covered)

    return {'Covered': covered, 'ExpValue': round(ev, 2), 'WinProb': round(win, 2), 'AvgOdds': round(avg_odds, 2)}

def elimination(lst):
    cond = [~pl.col('Covered').list.contains(l) for l in lst]
    return pl.all_horizontal(cond)


def format_trade(trade):
    return {'Banker': trade.get('Banker',None),
            'Covered': trade['Covered'],
            'WinProb': round(trade['WinProb'], 2),
            'ExpValue': round(trade['ExpValue'], 2),
            'AvgOdds': round(trade['AvgOdds'], 2)}


@app.route('/qprec/<int:race_no>/<int:banker>', methods=['GET'])
def disp_qp_recs(race_no=1, banker=1, exclude=[], maxC=None):
    use_filter = request.args.get('filter')
    use_filter = use_filter and (use_filter.lower() == 'true')
    exclude = request.args.get('exclude')
    if exclude:
        excluded = [int(v) for v in exclude.split(',')]
    maxC = request.args.get('maxC', type=int)

    df_trades = generate_all_qp_trades(
        '', '', race_no, fit_harville=use_filter)
    df_trades = df_trades.filter(pl.col('Banker') == banker)
    df_trades = df_trades.filter(pl.col('ExpValue') >= 0.05)
    if exclude:
        df_trades = df_trades.filter(elimination(excluded))
    if maxC:
        df_trades = df_trades.filter(pl.col('NumCovered') <= maxC)
    pareto_trades = pareto_filter(df_trades, ['NumCovered'], ['WinProb', 'ExpValue']).sort(
        'WinProb', descending=True)

    return [format_trade(t) for t in pareto_trades.iter_rows(named=True)]




@app.route('/plarec/<int:race_no>', methods=['GET'])
def disp_pla_recs(race_no=1, exclude=[], maxC=None):
    exclude = request.args.get('exclude')
    if exclude:
        excluded = [int(v) for v in exclude.split(',')]
    maxC = request.args.get('maxC', type=int)

    df_trades = generate_all_pla_trades('', '', race_no)
    df_trades = df_trades.filter(pl.col('ExpValue') >= 0.05)
    if exclude:
        df_trades = df_trades.filter(elimination(excluded))
    if maxC:
        df_trades = df_trades.filter(pl.col('NumCovered') <= maxC)
    pareto_trades = pareto_filter(df_trades, ['NumCovered'], ['WinProb', 'ExpValue']).sort(
        'WinProb', descending=True)

    return [format_trade(t) for t in pareto_trades.iter_rows(named=True)]



"""Functions to fetch and process historical race and horse data from HKJC
"""
from __future__ import annotations

import requests
import polars as pl
from bs4 import BeautifulSoup
from cachetools.func import ttl_cache

from .utils import _parse_html_table

HKJC_RACE_URL_TEMPLATE = "https://racing.hkjc.com/racing/information/English/Racing/LocalResults.aspx?RaceDate={date}&Racecourse={venue_code}&RaceNo={race_number}"
HKJC_HORSE_URL_TEMPLATE = "https://racing.hkjc.com/racing/information/English/Horse/Horse.aspx?HorseNo={horse_no}"

incidents = ['DISQ', 'DNF', 'FE', 'ML', 'PU', 'TNP', 'TO',
             'UR', 'VOID', 'WR', 'WV', 'WV-A', 'WX', 'WX-A', 'WXNR']

REQUEST_TIMEOUT = 10

HTML_HEADERS = {
    "Origin": "https://racing.hkjc.com",
    "Referer": "https://racing.hkjc.com",
    "Content-Type": "text/plain",
    "Accept": "*/*",
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/141.0.0.0 Safari/537.36",
}


@ttl_cache(maxsize=100, ttl=3600)
def _soupify(url: str) -> BeautifulSoup:
    """Fetch and parse a webpage and return BeautifulSoup object
    """
    response = requests.get(url, timeout=REQUEST_TIMEOUT, headers=HTML_HEADERS)
    response.raise_for_status()
    return BeautifulSoup(response.content, 'html.parser')


def _soupify_race_page(date: str, venue_code: str, race_number: int) -> BeautifulSoup:
    """Fetch and parse HKJC race results page and return BeautifulSoup object
    """
    url = HKJC_RACE_URL_TEMPLATE.format(
        date=date, venue_code=venue_code, race_number=race_number)
    return _soupify(url)


def _soupify_horse_page(horse_no: str) -> BeautifulSoup:
    """Fetch and parse HKJC race results page and return BeautifulSoup object
    """
    url = HKJC_HORSE_URL_TEMPLATE.format(horse_no=horse_no)
    return _soupify(url)


def _classify_running_style(df: pl.DataFrame, running_pos_col="RunningPosition") -> pl.DataFrame:
    """Classify running style based on RunningPosition column
    """
    if df.height == 0:
        return df

    # Split the RunningPosition column into separate columns and convert to integers
    df = df.with_columns(
        pl.col(running_pos_col)
        .str.split_exact(" ", n=3)
        .struct.rename_fields(["StartPosition", "Position2", "Position3", "FinishPosition"])
        # Give an alias to the struct for easier selection
        .alias("split_data").cast(pl.Int64, strict=False)
    ).unnest("split_data")

    df = df.with_columns(
        pl.col('FinishPosition').fill_null(pl.col('Position3')))

    df = df.with_columns([
        (pl.col("StartPosition")-pl.col("FinishPosition")).alias("PositionChange"),
        pl.mean_horizontal("StartPosition", "Position2").alias(
            "AvgStartPosition"),
    ]).with_columns(pl.when(pl.col("StartPosition").is_null()).then(pl.lit("--"))
                    .when((pl.col("AvgStartPosition") <= 3) & (pl.col("StartPosition") <= 3)).then(pl.lit("FrontRunner"))
                    .when((pl.col("PositionChange") >= 1) & (pl.col("StartPosition") >= 6)).then(pl.lit("Closer"))
                    .otherwise(pl.lit("Pacer")).alias("RunningStyle"))

    recent_style = df['RunningStyle'][:5].mode()[0]
    df = df.with_columns(pl.lit(recent_style).alias("FavoriteRunningStyle"))

    return df


def _extract_horse_data(horse_no: str) -> pl.DataFrame:
    """Extract horse info and history from horse page
    """
    soup = _soupify_horse_page(horse_no)
    horse_name = soup.find('title').get_text().split('- Horses -')[0].strip()
    table = soup.find('table', class_='bigborder')
    horse_data = _parse_html_table(table).filter(
        pl.col('Date') != '')  # Remove empty rows
    if horse_data.height > 0:
        horse_data = _classify_running_style(horse_data)
        horse_data = horse_data.with_columns([
            pl.lit(horse_no).alias('HorseNo'),
            pl.lit(horse_name).alias('HorseName')
        ])

    return horse_data


def _clean_horse_data(df: pl.DataFrame) -> pl.DataFrame:
    """ Clean and convert horse data to suitable data types
    """
    if df.height == 0:
        return df

    df = df.with_columns(
        pl.col('Pla').str.split(' ').list.first().alias('Pla')
    ).filter(~pl.col('Pla').is_in(incidents))

    df = df.with_columns([
        pl.col('Pla').cast(pl.Int64, strict=False),
        pl.col('ActWt').cast(pl.Int64, strict=False),
        pl.col('DeclarHorseWt').cast(pl.Int64, strict=False),
        pl.col('Dr').cast(pl.Int64, strict=False),
        pl.col('Rtg').cast(pl.Int64, strict=False),
        pl.col('Dist').cast(pl.Int64, strict=False),
        pl.col('WinOdds').cast(pl.Float64, strict=False)
    ])

    df = (df.filter(~pl.col('FinishTime').str.starts_with('--'))
          .with_columns(
        (
            pl.col("FinishTime").str.splitn(".", 2).struct.field("field_0").cast(pl.Int64) * 60 +
            pl.col("FinishTime").str.splitn(
                ".", 2).struct.field("field_1").cast(pl.Float64)
        ).cast(pl.Float64).round(2).alias("FinishTime")
    ))

    df = df.with_columns(
        pl.col('RCTrackCourse').str.split_exact(' / ', 2)
        .struct.rename_fields(['Venue', 'Track', 'Course'])
        .alias('RCTrackCourse')
    ).unnest('RCTrackCourse')

    df = df.with_columns(
        pl.when(pl.col('Date').str.len_chars() <= 8)
        .then(pl.col('Date').str.strptime(pl.Date, '%d/%m/%y', strict=False))
        .otherwise(pl.col('Date').str.strptime(pl.Date, '%d/%m/%Y'))
    ).with_columns(
        pl.concat_str(pl.col('Date').dt.strftime('%Y%m%d'), pl.col(
            'Venue'), pl.col('RaceIndex')).alias('RaceId')
    ).drop("VideoReplay")
    return df


def get_horse_data(horse_no: str) -> pl.DataFrame:
    df = _extract_horse_data(horse_no)
    return _clean_horse_data(df)


def _clean_race_data(df: pl.DataFrame) -> pl.DataFrame:
    """ Clean and convert horse data to suitable data types
    """
    if df.height == 0:
        return df

    df = df.with_columns(
        pl.col('Pla').str.split(' ').list.first().alias('Pla')
    ).filter(~pl.col('Pla').is_in(incidents))

    df = df.with_columns([
        pl.col('Pla').cast(pl.Int64, strict=False),
        pl.col('HorseNo').cast(pl.Int64, strict=False),
        pl.col('ActWt').cast(pl.Int64, strict=False),
        pl.col('DeclarHorseWt').cast(pl.Int64, strict=False),
        pl.col('Dr').cast(pl.Int64, strict=False),
        pl.col('WinOdds').cast(pl.Float64, strict=False)
    ])

    df = df.with_columns(
        (
            pl.col("FinishTime").str.splitn(":", 2).struct.field("field_0").cast(pl.Int64) * 60 +
            pl.col("FinishTime").str.splitn(
                ":", 2).struct.field("field_1").cast(pl.Float64)
        ).cast(pl.Float64).round(2).alias("FinishTime")
    )

    return df


def _extract_race_data(date: str, venue_code: str, race_number: int) -> pl.DataFrame:
    soup = _soupify_race_page(date, venue_code, race_number)
    table = soup.find('div', class_='race_tab').find('table')
    race_data = _parse_html_table(table)

    # Extract the relevant race information
    race_id = race_data.columns[0].replace(f'RACE{race_number}', '')
    race_class = race_data.item(1, 0).split('-')[0].strip()
    race_dist = race_data.item(1, 0).split('-')[1].strip().rstrip('M')
    race_name = race_data.item(2, 0).strip()
    going = race_data.item(1, 2).strip()
    course = race_data.item(2, 2).strip()

    race_info = {'Date': date,
                 'Venue': venue_code,
                 'RaceIndex': int(race_id),
                 'RaceNumber': race_number,
                 'RaceClass': race_class,
                 'RaceDistance': race_dist,
                 'RaceName': race_name,
                 'Going': going,
                 'Course': course}

    # Extract the results table
    table = soup.find('div', class_='performance').find('table')
    race_data = (_parse_html_table(table)
                 .with_columns([
                     pl.lit(value).alias(key) for key, value in race_info.items()
                 ])
                 .with_columns(
                     pl.col("Horse").str.extract(r"\((.*?)\)")
                     .alias("HorseID")
    )
    )

    return race_data


def get_race_data(date: str, venue_code: str, race_number: int) -> pl.DataFrame:
    df = _extract_race_data(date, venue_code, race_number)
    return _clean_race_data(df)



"""Top-level package for hkjc tools.

This module re-exports commonly used symbols from the submodules.
"""
from importlib.metadata import version as _version

__all__ = ["live", "features",
           "generate_all_qp_trades", "generate_all_pla_trades", "pareto_filter",
           "speedpro_energy", "speedmap", "harveille_model",
           "generate_historical_data"]

try:
    __version__ = _version(__name__)
except Exception:  # pragma: no cover - best-effort version resolution
    __version__ = "0.0.0"

from .processing import generate_all_qp_trades, generate_all_pla_trades, generate_historical_data
from .utils import pareto_filter
from .speedpro import speedmap, speedpro_energy
from . import harville_model, live, features
